(* $Id: util.ml,v 1.11 2008/01/23 10:41:08 deraugla Exp $ *)

open Font;
open Rtdecl;
open Std;
open Xlib;

type glob_struct = { xgcv : xGCValues; xev : xEvent; fds : fd_set };

type attribute_values =
  { backg_att : option paint;
    band_att : option int;
    bd_backg_att : option paint;
    bold_att : option int;
    border_att : option int;
    font_att : array font;
    foreg_att : option paint;
    height_att : option int;
    inter_att : option int;
    left_justif : bool;
    name_att : option string;
    width_att : option int }
;

value dynamo_tag (a : 'a) b r =
  (fun x (y : 'a) -> do { r.val := Some x; b },
   fun f -> do {
     r.val := None;
     let v = f a in
     match r.val with
     [ None -> failwith ("bad type \"" ^ v ^ "\", should be \"" ^ b ^ "\"")
     | Some x -> x ]
   })
;

value dynamo_local_info x = dynamo_tag C_LI x
and dynamo_global_info x = dynamo_tag C_GI x
and dynamo_args_info x = dynamo_tag C_AI x;

value no_info _ = "";
value no_user_info = C_UI (fun _ -> failwith "no_user_info");
value no_args_info _ = failwith "no_args_info";

value mallocated_var alloc_fun r _ =
  match r.val with
  [ None -> do {
      let v = alloc_fun () in r.val := Some v;
      v
    }
  | Some v -> v ]
;

value gstr =
  let gstr = ref None in
  fun () ->
    match gstr.val with
    [ Some gstr -> gstr
    | None -> do {
        let str =
          {xgcv = alloc_XGCValues (); xev = alloc_XEvent ();
           fds = alloc_fd_set ()}
        in
        gstr.val := Some str;
        str
      } ]
;

value create_widget xd win istop x y wsh wdesc info children =
  {wid_xd = xd; win = win; is_toplevel = istop; x = x; y = y;
   width = wsh.sh_width; height = wsh.sh_height; border = wsh.sh_border;
   wdesc = wdesc; is_mapped = False; frozen = False; deleted = False;
   delete_callback = None; info = info; user_info = no_user_info;
   children = children}
;

value add_widget name_att win wid = do {
  let xd = wid.wid_xd in hash_add_assoc (win, wid) xd.wid_by_win;
  match name_att with
  [ Some wname -> hash_add_assoc (wname, wid) xd.wid_by_name
  | _ -> () ];
  wid
}
and remove_widget name_att win wid = do {
  let xd = wid.wid_xd in
  match name_att with
  [ Some wname -> hash_remove_assoc wname xd.wid_by_name
  | _ -> () ];
  wid.deleted := True;
  hash_remove_assoc win xd.wid_by_win
}
and add_ginfo xd wname global_info ginfo = do {
  hash_add_assoc (wname, global_info ginfo) xd.ginfo;
  ginfo
}
and ginfo xd wname = hash_assoc wname xd.ginfo
and remove_ginfo xd wname = hash_remove_assoc wname xd.ginfo;

value opt_val def =
  fun
  [ Some x -> x
  | None -> def ]
;

value create_window xd pwin is_top x y wsh att_val emask = do {
  let {sh_width = width; sh_height = height; sh_border = border} = wsh in
  let bg_att = att_val.backg_att in
  let bd_att = att_val.bd_backg_att in
  let bg =
    match bg_att with
    [ Some (ColorPn c) -> c.pixel
    | _ -> xd.white ]
  in
  let bd =
    match bd_att with
    [ Some (ColorPn c) -> c.pixel
    | _ -> xd.black ]
  in
  let win =
    xCreateSimpleWindow
      (xd.dpy, pwin, x, y, max width 1, max height 1, max border 0, bd, bg)
  in
  match bg_att with
  [ Some NonePn -> xSetWindowBackgroundPixmap (xd.dpy, win, xNone)
  | Some (PixmapPn p) -> xSetWindowBackgroundPixmap (xd.dpy, win, p.pixmap)
  | _ -> () ];
  match bd_att with
  [ Some NonePn -> xSetWindowBorderPixmap (xd.dpy, win, xNone)
  | Some (PixmapPn p) -> xSetWindowBorderPixmap (xd.dpy, win, p.pixmap)
  | _ -> () ];
  let emask =
    if is_top then
      emask lor focusChangeMask lor keyPressMask lor enterWindowMask lor
      leaveWindowMask
    else emask
  in
  xSelectInput (xd.dpy, win, emask);
  win
};

value set_wm_and_class_hints =
  let xwmh = mallocated_var alloc_XWMHints (ref None) in
  fun wid -> do {
    let xd = wid.wid_xd
    and xwmh = xwmh () in
    set_XWMHints_input (1, xwmh);
    set_XWMHints_flags (inputHint, xwmh);
    xSetWMHints (xd.dpy, wid.win, xwmh);
    xSetClassHint (xd.dpy, wid.win, "camlrt", "Camlrt")
  }
;

value atoms = alloc_XAtom 1;

value set_delete_callback wid wdel =
  let xd = wid.wid_xd in
  match wdel with
  [ Some callb -> do {
      set_XAtom (xd.wm_delete_window, atoms, 0);
      let _ : status = xSetWMProtocols (xd.dpy, wid.win, atoms, 1) in
      wid.delete_callback := Some callb
    }
  | None -> () ]
;

value set_std_prop =
  let xsh = mallocated_var alloc_XSizeHints (ref None) in
  fun wid wname iname x y wsh flags -> do {
    let xd = wid.wid_xd
    and xsh = xsh () in
    set_XSizeHints_x (x, xsh);
    set_XSizeHints_y (y, xsh);
    set_XSizeHints_width (wsh.sh_width, xsh);
    set_XSizeHints_height (wsh.sh_height, xsh);
    let flags =
      if wsh.width_inc == -1 && wsh.height_inc == -1 then flags
      else do {
        (* xsh.width must be xsh.base_width + k * xsh.width_inc for some k.
           In other words, xsh.width and xsh.base_width must have the same
           rest of the division by xsh.width_inc.
           (same thing for height)
           Otherwise the Window Manager (or the X server, I don't know),
           may generate a ConfigureNotify to change the size! *)
        let width_inc = max 1 wsh.width_inc in
        let height_inc = max 1 wsh.height_inc in
        let min_width =
          (* = wsh.base_width + r, so that this sum has the same rest by
             division by width_inc as wsh.sh_width *)
          wsh.sh_width -
          (wsh.sh_width - wsh.base_width) / width_inc * width_inc
        in
        let min_height =
          (* same comment as for "min_width" above *)
          wsh.sh_height -
          (wsh.sh_height - wsh.base_height) / height_inc * height_inc
        in
        set_XSizeHints_min_width (min_width, xsh);
        set_XSizeHints_min_height (min_height, xsh);
        set_XSizeHints_width_inc (width_inc, xsh);
        set_XSizeHints_height_inc (height_inc, xsh);
        flags lor pMinSize lor pResizeInc
      }
    in
    set_XSizeHints_flags (flags, xsh);
    xSetStandardProperties
      (xd.dpy, wid.win, wname, iname, xNone, Sys.argv.(0), xsh)
  }
;

value lnot i = -i - 1;

value make_array len f = do {
  let v = Array.create len (Obj.magic 0) in
  for i = 0 to len - 1 do { v.(i) := f i };
  v
};

value item_modified state item =
  {shiftMod = state land shiftMask != 0; shiftLock = state land lockMask != 0;
   control = state land controlMask != 0; mod1 = state land mod1Mask != 0;
   item = item}
;

value attribute_values =
  let backg_att = ref None in
  let band_att = ref None in
  let bd_backg_att = ref None in
  let bold_att = ref None in
  let border_att = ref None in
  let font_att = ref [| |] in
  let foreg_att = ref None in
  let height_att = ref None in
  let inter_att = ref None in
  let left_justif = ref False in
  let name_att = ref None in
  let width_att = ref None in
  fun attr -> do {
    List.iter
      (fun
       [ BackgroundAtt v -> backg_att.val := Some v
       | BandAtt v -> band_att.val := Some v
       | BoldAtt v -> bold_att.val := Some v
       | BorderBackgAtt v -> bd_backg_att.val := Some v
       | BorderAtt v -> border_att.val := Some v
       | FontAtt v -> font_att.val := v
       | ForegroundAtt v -> foreg_att.val := Some v
       | HeightAtt v -> height_att.val := Some v
       | InterAtt v -> inter_att.val := Some v
       | LeftJustifAtt -> left_justif.val := True
       | NameAtt v -> name_att.val := Some v
       | WidthAtt v -> width_att.val := Some v
       | _ -> () ])
      attr;
    let r =
      {backg_att = backg_att.val; band_att = band_att.val;
       bd_backg_att = bd_backg_att.val; border_att = border_att.val;
       bold_att = bold_att.val; font_att = font_att.val;
       foreg_att = foreg_att.val; height_att = height_att.val;
       inter_att = inter_att.val; left_justif = left_justif.val;
       name_att = name_att.val; width_att = width_att.val}
    in
    backg_att.val := None;
    band_att.val := None;
    bd_backg_att.val := None;
    border_att.val := None;
    bold_att.val := None;
    font_att.val := [| |];
    foreg_att.val := None;
    height_att.val := None;
    inter_att.val := None;
    left_justif.val := False;
    name_att.val := None;
    width_att.val := None;
    r
  }
;

value nround x y =
  if x > 0 && y > 0 || x < 0 && y < 0 then (2 * x + y) / (2 * y)
  else (2 * x - y) / (2 * y)
;

value mX = 255;

value hsv_of_rgb (r, g, b) =
  let r = max 0 (min mX r) in
  let g = max 0 (min mX g) in
  let b = max 0 (min mX b) in
  let v = max r (max g b)
  and x = min r (min g b) in
  (if v = 0 || v = x then 0
   else
     let r0 = (v - r) * mX / (v - x)
     and g0 = (v - g) * mX / (v - x)
     and b0 = (v - b) * mX / (v - x) in
     let h =
       if r = v then if g = x then 5 * mX + b0 else mX - g0
       else if g = v then if b = x then mX + r0 else 3 * mX - b0
       else if r = x then 3 * mX + g0
       else 5 * mX - r0
     in
     h / 6,
   if v = 0 then 0 else nround (mX * (v - x)) v, v)
;

value rgb_of_hsv (h, s, v) =
  let h = max 0 (min mX h) in
  let s = max 0 (min mX s) in
  let v = max 0 (min mX v) in
  let h = h * 6 in
  let i = h / mX * mX in
  let f = h - i in
  let m = v * (mX - s) / mX
  and n = v * (mX - s * f / mX) / mX
  and k = v * (mX - s * (mX - f) / mX) / mX in
  (nround
     (mX *
      (match i / mX with
       [ 0 | 6 -> v
       | 1 -> n
       | 2 -> m
       | 3 -> m
       | 4 -> k
       | 5 -> v
       | _ -> failwith "red_of_hsv" ]))
     mX,
   nround
     (mX *
      (match i / mX with
       [ 0 | 6 -> k
       | 1 -> v
       | 2 -> v
       | 3 -> n
       | 4 -> m
       | 5 -> m
       | _ -> failwith "green_of_hsv" ]))
     mX,
   nround
     (mX *
      (match i / mX with
       [ 0 | 6 -> m
       | 1 -> m
       | 2 -> k
       | 3 -> v
       | 4 -> v
       | 5 -> n
       | _ -> failwith "blue_of_hsv" ]))
     mX)
;

value get_default dpy str dval =
  let v = xGetDefault (dpy, "rt", str) in
  if is_null_C_String v then dval
  else string_of_C_String (v, c_String_length v)
;
